#
#  Copyright (c) 1998, 1999, Digital Creations, Fredericksburg, VA, USA.  
#  All rights reserved.
#  
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#  
#    o Redistributions of source code must retain the above copyright
#      notice, this list of conditions, and the disclaimer that follows.
#  
#    o Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions, and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#  
#    o Neither the name of Digital Creations nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#  
#  
#  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
#  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
#  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
#  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
#  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
#  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
#  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#  DAMAGE.
#
#  $Id: ociCurs.py,v 1.8 1999/08/27 18:11:31 petrilli Exp $
#
#  If you have questions regarding this software,
#  contact:
# 
#    Digital Creations L.C.  
#    info@digicool.com
# 
#    (540) 371-6909
#
##############################################################################

"""DBI-Compliant Oracle Database Interface
"""

jimwashere=0

import oci_, dbi, ociProc

try: from oci_8 import lob
except: lob = None

from ocitypes import *

from Buffer import Buffer
import ociBind
from ociUtil import RowID, oDate, error
from string import join
try:
    import Missing
    MissingValue=Missing.Value
except:
    MissingValue=None


class Cursor:
    "Oracle Database Cursors"
    arraysize=1
    description=None
    _iarraysize=10
    _s=_out=None
    

    def __init__(self, oci_connection):
        self._c=oci_.newCursor(oci_connection)

    def _error(self):
        rc=oci_.cda_def_rc_get(oci_.cucda(self._c))
        raise error, (rc, oci_.OracleErrorMessage(self._c.lda, rc))

    def _describe(self, ft=4):
        out=[]
        c=self._c
        if ft != 4:
            self._out=out
            return out # non-Select
        desc=[]
        i=1
        
        arraysize=self.arraysize
        if arraysize==1: arraysize=self._iarraysize
        else: self._iarraysize=arraysize
        
        odesc=[]
        while 1:
            d=oci_.odescr(c, i)
            status=d[0]
            if status:
                rc=oci_.cda_def_rc_get(oci_.cucda(c))
                if rc!=1007: self._error()
                break
            odesc.append(d)
            dbtype=d[2]
            if dbtype in (OCI_INTERNAL_LONG, OCI_INTERNAL_LONG_RAW,
                          OCI_INTERNAL_CLOB, OCI_INTERNAL_BLOB):
                arraysize=1
            i=i+1

        i=1
        for status, dbsize, dbtype, name, dsize, prec, scale, nullok in odesc:
            if have_dbi_type_for(dbtype): dbitype=dbitypes[dbtype]
            else: dbitype=dbtype
            desc.append(
                (name, dbitype, dsize, dbsize, prec, scale, nullok))
            ind=Buffer(arraysize,'h')
            alen=Buffer(arraysize,'H')
            arcode=Buffer(arraysize,'H')
            if dbtype in (OCI_INTERNAL_VARCHAR, OCI_INTERNAL_RAW, OCI_INTERNAL_CHAR):
                # String
                buf=Buffer(dbsize*arraysize)
                f=1
            elif dbtype==OCI_INTERNAL_LONG or dbtype==OCI_INTERNAL_LONG_RAW:
                # Use piecewise fetch for long columns"
                buf=([],Buffer(8192),Buffer(1,'I'))
                buf[-1][0]=8192
                oci_.definpw(c,i,buf,1000000000,dbtype,0,ind,alen,arcode)
                out.append(buf,dbsize,dbtype,ind,alen,arcode, 2)
                i=i+1
                continue
            elif dbtype==OCI_INTERNAL_NUMBER:
                if jimwashere:
                    buf=Buffer(1,21)
                    dbtype=OCI_EXTERNAL_VARNUM
                    dbsize=21
                    f=None
                else:
                    if scale:
                        buf=Buffer(arraysize,'d')
                        dbtype=OCI_EXTERNAL_FLOAT
                        dbsize=8
                    else:
                        buf=Buffer(arraysize,'i')
                        dbtype=OCI_EXTERNAL_INTEGER
                        dbsize=4
                    f=None
            elif dbtype==OCI_INTERNAL_ROWID:
                buf=Buffer(arraysize,11)
                f=RowID
            elif dbtype==OCI_INTERNAL_DATE: # Date
                buf=Buffer(arraysize,7)
                f=oDate
            elif dbtype==OCI_INTERNAL_BLOB or dbtype==OCI_INTERNAL_CLOB:
                if lob is None:
                    raise error, (
                        "Invalid LOB type code, %s, for column %s. "
                        "LOBs are not supported."
                        % (dbtype, i))
                buf=lob(c.lda)
                if buf.define(c,i,dbtype,ind,alen,arcode): self._error()
                out.append(buf,None,dbtype,ind,alen,arcode, 3)
                i=i+1
                continue
            else:
                raise error, (
                    "Invalid type code, %s, for column %s"
                    % (dbtype, i))
            if oci_.define(c,i,buf,dbsize,dbtype,ind,alen,arcode):
                self._error()
            out.append(buf,dbsize,dbtype,ind,alen,arcode, f)
            i=i+1

        if desc:
            self.description=tuple(desc)
            self._out=tuple(out)
            self._arraysize=arraysize
            self._as=-1
            self._i=-1
            self._ai=0
            self._out=out, [None]*len(out), range(len(out)), arraysize
        else:
            if self.__dict__.has_key('description'): del self.description
            self._out=None

        return out

    def _parse(self, sql__):
        _cursor_sql = sql__             # Stuff it somewhere to look at!
        if oci_.oparse(self._c, sql__, 0, 1):
            self._error()
        self._ft = oci_.cda_def_ft_get(oci_.cucda(self._c))

    def _execute(self, params__=(), kw={}):
        c=self._c
        
        if params__ or kw:
            if type(params__) is not type(()): params__=params__, 
            params=[]
            i=0
            for p in params__:
                i=i+1
                params.append((":p%d" % i, p))
            for k, p in kw.items():
                params.append((':'+k, p))
            params.sort()
        else: params=params__
        
        if self.__dict__.has_key('description'): del self.description
        _parray, self._cruft = ociBind.bindParams(self, params)
        out=self._describe(self._ft)

        if out:
            if _parray and _parray > 1:
                raise error, 'Cannot use array input with queries'
            arraysize=self._arraysize
            if arraysize==1:
                if oci_.oexec(c): self._error()
            else:
                if oci_.oexfet(c,arraysize,0,0):
                    rc=oci_.cda_def_rc_get(oci_.cucda(self._c))
                    if rc!=1403: self._error()
                    self._as=oci_.cda_def_rpc_get(oci_.cucda(c))
                else: self._as=arraysize
        else:
            self._out=None
            if _parray==1 or _parray is None:
                if oci_.oexec(c):
                    rc=oci_.cda_def_rc_get(oci_.cucda(self._c))
                    if rc!=3129: self._error()
                    # Piecewise insert/update
                    while 1:
                        (status, piece, (v, b, bl), iter, index
                         ) = oci_.ogetpi(c)
                        if status: self._error()
                        l=len(v)
                        i=bl[1]
                        ll=l-i
                        if ll <= 8192: piece=oci_.OCI_LAST_PIECE
                        else: ll=8192
                        bl[0]=ll
                        bl[1]=ii=i+ll
                        b[:ll]=v[i:ii]
                        if oci_.osetpi(c,piece,b,bl): self._error()
                        if not oci_.oexec(c): break
                        rc=oci_.cda_def_rc_get(oci_.cucda(self._c))
                        if rc!=3129: self._error()
            elif oci_.oexn(c,_parray,0): self._error()
            if _dml(self._ft): return 1 # Why oh why!
            return oci_.cda_def_rpc_get(oci_.cucda(c))

    def __getitem__(self, i):
        _i=self._i
        if i==_i: return self._r
        if i < _i: raise error, ('Index out of sequence', i)
        f=self.fetchone
        while i > _i:
            r=f()
            if r is not None: _i=_i+1
            else:
                raise IndexError, i
        self._i=_i
        return r

    def __len__(self): return 1

    def execute(self, sql__, params__=(), **kw):
        self._parse(sql__)
        r = self._execute(params__, kw)
        return r
        
    def fetchone(self, many=None):
        out=self._out
        if out:
            ai=self._ai
            if ai is None: return None
            i=self._i
            as=self._as
            out, r, rng, arraysize = self._out
            if ai >= as:
                if as < arraysize and as >= 0:
                    self._ai=None
                    return None
                c=self._c
                if arraysize==1:
                    if oci_.ofetch(c):
                        rc=oci_.cda_def_rc_get(oci_.cucda(c))
                        if rc==3130:
                            (status, piece, (l, b, bl), iter, index,
                             ) = oci_.ogetpi(c)
                            bl[0] = len(b)
                            del l[:]
                            oci_.osetpi(c, piece, b, bl)
                            while 1:
                                (status, piece, (l, b, bl), iter, index
                                 ) = oci_.ogetpi(c)
                                oci_.osetpi(c,piece,b,bl)
                                if oci_.ofetch(c):
                                    rc=oci_.cda_def_rc_get(oci_.cucda(c))
                                    if rc!=3130: self._error()
                                    l.append(b[:bl[0]])
                                else:
                                    l.append(b[:bl[0]])
                                    break
                        elif rc==1403: # End of data
                            self._ai=None
                            return None
                        else:
                            self._error()
                    ai=0
                else:
                    if oci_.ofen(c, arraysize):
                        rc=oci_.cda_def_rc_get(oci_.cucda(c))
                        if rc!=1403: self._error()
                        as=oci_.cda_def_rpc_get(oci_.cucda(c))-i-1
                        if as <= 0:
                            self._ai=None
                            return None
                        self._as=as
                    else:
                        self._as=arraysize
                    ai=self._ai=0

            for io in rng:
                buf,dbsize,dbtype,ind,alen,arcode,f = out[io]
                if ind[ai] < 0: v=MissingValue
                elif f is None: v=buf[ai]
                elif f==1:
                    f=dbsize*ai
                    v=buf[f:f+alen[ai]]
                elif f==2: v=join(buf[0],'')
                elif f==3:
                    v=buf
                    if many: v=v.read()
                else: v=f(buf[ai])
                r[io]=v

            r=tuple(r)
            self._i=i+1
            self._ai=ai+1
            self._r=r
            return r

        raise error, 'not querying'

    def fetchmany(self, n=None):            
        if n is None:
            n=self.arraysize
            if n < 1: n=1
        elif n < 1: return
        r=[]
        f=self.fetchone
        i=0
        while 1:
            v=f(1)
            if not v: break
            r.append(v)
            i=i+1
            if i >= n: break
        return r

    def fetchall(self):
        r=[]
        f=self.fetchone
        while 1:
            v=f(1)
            if not v: break
            r.append(v)
        return r

    def close(self):
        # This one is insane.
        oci_.oclose(self._c)
        del self._c

    def setinputsizes(self, sizes): pass
    def setoutputsizes(self, sizes): pass

    # Support the procedures attribute:
    def __getattr__(self, name):
        if name=='procedures':
            p=self.procedures=ociProc.StoredProcedureNamespace(self)
            return p
        raise AttributeError, name

class Handle(Cursor):
    def execute(self, params__=(), **kw):
        r = self._execute(params__, kw)
        return r


# Set up function to check for DML statements to conform with the
# DBI spec.
_dml={}
for _i in ([1,2,6,7,8]+range(10,14+1)+range(16,33+1)+[35]+
           range(37,68+1)+range(70-76+1)):
    _dml[_i]=1
_dml=_dml.has_key

dbitypes={
    2: dbi.NUMBER,
    1: dbi.STRING, 
    8: dbi.STRING, 
    96: dbi.STRING, 
    112: dbi.STRING, 
    12: dbi.DATE, 
    23: dbi.RAW,
    24: dbi.RAW,
    113: dbi.RAW,
}
have_dbi_type_for=dbitypes.has_key
    
    
